#define PI 3.141592653589
#define USE_BILINEAR_PCF_TAP 0

#define SHADOW_FACTOR 3000 // max distance from 0,0,0 to any point (in direction of sun)
#define PCF_SIDE 2 // PCF taps form a grid
#define FAKE_PCF_SIDE 2 // FAKE_PCF taps form a grid within each cell of the PCF grid
#define SPREAD 0.005


// PCSS params
#define SHADOW_DIST1 20
#define SHADOW_DIST2 50
#define SHADOW_DIST3 200

// fade shadow out
#define SHADOW_FADE_START 150
#define SHADOW_FADE_END 200


#define SHADOW_BIAS1 2/(-2.11+32.19*max(0.1,light_exposure))
#define SHADOW_BIAS2 2/(-.537+10.355*max(0.1,light_exposure))
#define SHADOW_BIAS3 2/(-0.187+3.533*max(light_exposure,0.1))

#define DISCARD return

float4x4 world;
float4x4 view_proj;
float4x4 shadow_world_view_proj1;
float4x4 shadow_world_view_proj2;
float4x4 shadow_world_view_proj3;
float3 camera_pos;
float4 light_pos;
float4 light_pos_ws;
float3 light_ambient;
float3 surf_ambient;

#if USE_TEXTURE==1
        sampler2D diffuse_texture : TEXUNIT0;
        sampler2D shadow_texture1 : TEXUNIT1;
        sampler2D shadow_texture2 : TEXUNIT2;
        sampler2D shadow_texture3 : TEXUNIT3;
#else
        sampler2D shadow_texture1 : TEXUNIT0;
        sampler2D shadow_texture2 : TEXUNIT1;
        sampler2D shadow_texture3 : TEXUNIT2;
#endif
float4 the_fog_params;
float3 the_fog_colour;
float4 visibility;
float alpha_rej;
float3 light_diffuse;
float3 light_specular;
float3 surf_diffuse;
float3 surf_specular;
float surf_shininess;


void vp_main (
        in float3 vertex : POSITION,
        in float3 normal : NORMAL,
#if USE_TEXTURE==1
        in float2 uv0 : TEXCOORD0,
#endif
#if USE_VERTEX_COLOURS==1
        in float4 colour : COLOR,
#if USE_NIGHT_COLOURS==1
        in float4 uv1 : TEXCOORD1,
#endif
#endif
#if USE_TEXTURE==1
        out float2 uv0_ : TEXCOORD0,
#endif
        out float4 vertex_ls1_ : TEXCOORD1,
        out float4 vertex_ls2_ : TEXCOORD2,
        out float4 vertex_ls3_ : TEXCOORD3,
        //out float4 vertex_ls4_ : POSITION4,
        out float3 vertex_to_camera_ : TEXCOORD4,
        out float3 vertex_to_light_ : TEXCOORD5,
        out float3 normal_ : TEXCOORD6,
        out float3 dyn_light_ : COLOR1,
        out float2 ffs_ : TEXCOORD7,

        out float4 gl_Position : POSITION,
        out float4 gl_FrontColor : COLOR0
) {
        #if USE_VERTEX_COLOURS==1
        dyn_light_ = colour.rgb;
        gl_FrontColor.a = colour.a;
        #else
        dyn_light_ = float3(0,0,0);
        gl_FrontColor.a = 1;
        #endif

        // do ambient lighting
        gl_FrontColor.rgb = light_ambient * surf_ambient;

        // lighting vectors for fragment program
        vertex_to_light_ = normalize(light_pos.xyz - vertex*light_pos.w);
        vertex_to_camera_ = normalize(camera_pos - vertex);
        normal_ = normal;

        #if USE_TEXTURE==1
        uv0_ = uv0;
        #endif

        // vertex position in light space for shadow mapping
        vertex_ls1_ = mul(shadow_world_view_proj1, float4(vertex,1));
        vertex_ls2_ = mul(shadow_world_view_proj2, float4(vertex,1));
        vertex_ls3_ = mul(shadow_world_view_proj3, float4(vertex,1));

        float3 vertex_ws = mul(world, float4(vertex,1)).xyz;
        ffs_.y = -dot(vertex_ws,light_pos_ws.xyz);

        // for the rasteriser
        gl_Position = mul(view_proj, float4(vertex_ws,1));
        ffs_.x = gl_Position.z;
}

float tap_dist (sampler2D tex, float2 suv, float our_dist)
{
        return our_dist - tex2D(tex, suv).r;
}

float tap_test (sampler2D tex, float2 suv, float our_dist)
{
        return tap_dist(tex,suv,our_dist)>0 ? 1.0 : 0.0;
}

float bilinear_tap_test (sampler2D tex, int sz, float2 suv, float d)
{
        #if USE_BILINEAR_PCF_TAP == 0
        return tap_test(tex,suv,d);
        #else
        suv *= sz;
        float2 wh = floor(suv); // whole component
        float2 fr = suv - wh; // fractional component
        return lerp(lerp(tap_test(tex,(wh+float2(0,0))/sz,d), tap_test(tex,(wh+float2(0,1))/sz,d), fr.y),
                    lerp(tap_test(tex,(wh+float2(1,0))/sz,d), tap_test(tex,(wh+float2(1,1))/sz,d), fr.y),
                    fr.x);
        #endif
}

float test_shadow (float4 vertex_ls, sampler2D tex, float our_dist,
                   float bias, float2 screen_pos)
{
        if (vertex_ls.x < 0) return 0;
        if (vertex_ls.y < 0) return 0;
        if (vertex_ls.x > vertex_ls.w) return 0;
        if (vertex_ls.y > vertex_ls.w) return 0;

        our_dist -= bias;
        our_dist /= SHADOW_FACTOR;

        float spread = SPREAD;


        // range (0,1) on each axis, equal border around all edges
        float2 offset = (float2(int(screen_pos.x)%FAKE_PCF_SIDE+0.5,
                                int(screen_pos.y)%FAKE_PCF_SIDE+0.5)) / FAKE_PCF_SIDE;

        if (offset.x > 0.5) offset.y = 1 - offset.y;

        float total = 0;
        for (int y=0 ; y<PCF_SIDE ; ++y) {
                for (int x=0 ; x<PCF_SIDE ; ++x) {
                        float2 p = (float2(x,y) + offset)/PCF_SIDE - float2(0.5,0.5);
                        p *= spread;
                        float2 suv = (vertex_ls.xy/vertex_ls.w + p);
                        total += bilinear_tap_test(tex,1024,suv,our_dist);
                }
        }

        return total / PCF_SIDE / PCF_SIDE;
}


void fp_main (
        in float4 gl_Color : COLOR0,
#if USE_TEXTURE==1
        in float2 uv0_ : TEXCOORD0,
#endif
        in float4 vertex_ls1_ : TEXCOORD1,
        in float4 vertex_ls2_ : TEXCOORD2,
        in float4 vertex_ls3_ : TEXCOORD3,
        //out float4 vertex_ls4_ : TEXCOORD1,
        in float3 vertex_to_camera_ : TEXCOORD4,
        in float3 vertex_to_light_ : TEXCOORD5,
        in float3 normal_ : TEXCOORD6,
        in float2 ffs_ : TEXCOORD7,
        in float2 gl_Position_ : VPOS,
        in float3 dyn_light_ : COLOR1,

        out float4 pixel : COLOR
) {

        float2 screen_pos = gl_Position_;
        float cam_dist = ffs_.x;
        float light_dist = ffs_.y;

        // start with vertex lighting, if any
        float3 dyn_light = dyn_light_;
        float3 spec_light = float3(0,0,0);

        // compute diffuse lighting
        float light_exposure = saturate(dot(vertex_to_light_,normal_));
        dyn_light += light_exposure * light_diffuse * surf_diffuse;

        // compute specular lighting
        float3 r = reflect(-vertex_to_light_,normal_);
        spec_light = pow(max(0.00001,dot(r,vertex_to_camera_)),20)
                     * light_specular * surf_specular;


        // compute shadow (reduces dynamic lighting if in shadow)
        float shadowyness = 0;
        if (cam_dist < SHADOW_DIST1) {
                shadowyness = test_shadow(vertex_ls1_, shadow_texture1, light_dist, SHADOW_BIAS1, screen_pos);
        } else if (cam_dist < SHADOW_DIST2) {
                shadowyness = test_shadow(vertex_ls2_, shadow_texture2, light_dist, SHADOW_BIAS2, screen_pos);
        } else if (cam_dist < SHADOW_DIST3) {
                shadowyness = test_shadow(vertex_ls3_, shadow_texture3, light_dist, SHADOW_BIAS3, screen_pos);
        }

        float fade = min(1, (SHADOW_FADE_END-cam_dist)/(SHADOW_FADE_END-SHADOW_FADE_START));
        shadowyness *= fade;

        // shadow affects dynamic and specular light
        dyn_light.rgb -= dyn_light.rgb * shadowyness;
        spec_light.rgb -= spec_light.rgb * shadowyness;


        // ambient as pased from vertex shader, with the dyn_light just computed
        // note that alpha in vertex colours comes from gl_Color, not dyn_light_
        pixel.rgb = gl_Color.rgb + dyn_light;
        pixel.a = gl_Color.a;

        #if USE_TEXTURE==1
        float4 texel = tex2D(diffuse_texture, uv0_);
        pixel *= texel;
        #endif

        pixel.rgb += spec_light;


        // cull pixels after lighting and texture, but before fading
        if (pixel.a <= alpha_rej) DISCARD;

        // exp2 fog
        float clarity = the_fog_params.x * cam_dist;
        clarity = clamp(exp(-clarity * clarity),0.0,1.0);
        pixel.rgb = lerp(the_fog_colour, pixel.rgb, clarity);

        pixel.a *= visibility.x;

}

// vim: ts=8:sw=8:et
