print("Loading world.lua")

World = World or { }

function World.new()
        local self = { }
        self.gom = get_gom()

        self.shootFast = 100
        self.shootSlow = 10
        self.shootSpin = Vector3(0,0,30)
        self.visibility = 1

        local function frameCallback()
                gfx.scnmgr:updateSceneGraph()
                local pos = gfx.cam.derivedPosition
                local vis = self.visibility
                vis = vis * clamp(1+pos.z/500,1,1.5)
                self.gom:centre(pos.x,pos.y,pos.z,vis)

        end

        main.frameCallbacks:insert("World.frameCallback",frameCallback)

        ui:bind("F3", function() self:clearTemporary() end)
        ui:bind("F4", function() self:clearAll() end)

        make_instance(self,World)

        return self

end

function World:camRay()
        return physics.world:cast(-1,gfx.cam.derivedPosition,
                                  gfx.cam.derivedOrientation * V_FORWARDS,
                                  8000, true, false)
end


function World:pickPos()
        local dist = self:camRay()
        if dist == nil then
                error("Not pointing at anything",2)
        end
        return gfx.cam.derivedPosition + gfx.cam.derivedOrientation*(dist*V_FORWARDS)
end

function World:allActivated()
        return filter(gom.objects,function(x) return x.activated end)
end

function World:bodyToObj(body)
        local o = filter(gom.objects,function(x) return x.activated and x.instance.body == body end)[1]
        return o
end

function World:pickObjSafe()
        local _, body = self:camRay()
        if body == nil then return nil end
        return self:bodyToObj(body)
end
function World:pickObj()
        local body = self:pickObjSafe()
        if body == nil then
                error("Not pointing at anything",2)
        end
        return body
end

function World:allOfType(class, startWith)
        startWith = startWith or gom.objects
        return filter(startWith,function(x) return x.className == class end)
end

function World:place(class,height,perm,pos,quat)
        height = height or 0
        perm = perm or true
        pos = pos or self:pickPos()
        quat = quat or gfx:yaw()
        local tab = {test = 1, rot=quat}
        if not perm then
                tab.killMe = 1
        end
        local x,y,z = pos:xyz()
        z = z + height
        self.gom:addObject(class,x,y,z,tab)
end

function World:clearTemporary()
        for _,v in ipairs(self.gom.objects) do
                if v.killMe then
                        v:destroy()
                elseif v.test then
                        v:deactivate() -- will reactivate immediately if within range
                end
        end
end


function World:deactivateAll()
        for _,v in ipairs(self.gom.objects) do
                v:deactivate()
        end
end


function World:clearAll()
        for _,v in ipairs(self.gom.objects) do
                if v.test then
                        v:destroy()
                end
        end
end


function World:fireExtended(v,t,spin)
        v = v or 40
        t = t or "Crate"
        local q = gfx.cam.derivedOrientation
        local x,y,z = gfx.cam.derivedPosition:xyz()
        local o = self.gom:addObject(t,x,y,z,{rot=q,killMe=1})
        o:activate()
        o.instance.body.linearVelocity = q * Vector3(0,v,0)
        if spin then
                o.instance.body.angularVelocity = q * spin
        end
end



function World:fire (type)
        local speed = ui:ctrl() and self.shootFast or self.shootSlow
        local spin = ui:alt() and self.shootSpin
        self:fireExtended(speed,type,spin)
end

function World:overrideClass (name, superclass, tab)
        superclass = superclass or {}
        tab = tab or {}
        if self.gom:hasClass(name) then
                local class = self.gom:getClass(name)
                for k,v in pairs(tab) do
                        class[k] = v
                end
                class.parent = superclass
        else
                self.gom:addClass(name, superclass, tab)
        end
end



function World:overrideObjectV (class,pos,tab)
        return self:overrideObject(class,pos.x,pos.y,pos.z,tab)
end

function World:overrideObject (class,x,y,z,tab)
        local name = tab.name
        if self.gom:hasObject(name) then
                self.gom:removeObject(name)
        end
        return self.gom:addObject(class,x,y,z,tab)
end


function World:printData (object)
        if object == nil then
                error("object was nil")
        end
        if not object.activated then
                error("object is not activated")
        end
        local class = object.className
        local x, y, z = object.instance.gfx.derivedPosition:xyz()
        local rot = object.instance.gfx.derivedOrientation
        echo("gom:addObject(\""..class.."\", "..x..", "..y..", "..z..", {rot="..tostring(rot).."})")
end


function World:destroy()
        main.frameCallbacks:removeByName("World.frameCallback")
end

if world ~= nil then
        world:destroy()
end
world = World.new()

-- vim: shiftwidth=8:tabstop=8:expandtab
