print("Loading friction_models.lua")


function global_friction_model (wheel, restitution, speed, drive_force, traction_control)

        local vehicle = wheel.obj
        -- first compute limiting friction
        -- this is the force required to keep the object stationary
        -- no matter how fast it is going
        local num_wheels = #vehicle.wheels
        local body = vehicle.body
        local world = vehicle.world
        local mass = body.mass / num_wheels
        local limiting_force = speed * mass / world.stepSize

        -- reduce the amount of play this force has in the eventual consensus
        limiting_force = limiting_force * traction_control

        -- now incorporate the drive force (at right angles)
        local required_force = math.sqrt(limiting_force * limiting_force + drive_force * drive_force)
        if required_force == 0 then
                required_force = 1 -- avoid NaN a few lines below
        end

        -- compute them maximum frictional force the surface can provide
        local mu = wheel.mu -- TODO: include road surface in mu
        local theoretical_max = mu * restitution

        --if theoretical_max < required_force then
        --        echo ("slip: "..required_force.." > "..theoretical_max)
        --end

        -- now we have the actual applied friction
        local applied_force = math.min(theoretical_max,required_force)

        -- split it back into the 2 components
        local applied_drive = drive_force  *  applied_force / required_force
        local applied_friction =  limiting_force  *  applied_force / required_force
        
        return applied_friction, applied_drive, traction_control < 0.5
end

-- vim: shiftwidth=8:tabstop=8:expandtab
